#!/usr/bin/perl -w
do "Utils.pl";
use Getopt::Long;
use strict;

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------

    my $infile = undef;         # path to .m3m input file
    my $outfile = undef;        # path to .m1s output file
    my @infiles = ();           # list of files to process
    my @includefiles = ();      # path to .vb include files
    my %includes = ();          # include file definitions
    my %macros = ();            # macro definitions
    my @activedefines = ();     # defines referenced in current file
    my $outdir = undef;         # default m1s file directory
    my $warnings = 0;
    my $help = 0;
    
    # Parse the command line
    GetOptions (
            'f=s'       => \$infile,
            'D=s'       => \$outdir,
            'h'         => \$help
    );
    
    # Was an explicit location given for the m1s files?
    if (! defined $outdir) {
        $outdir = "m1s";
    }
    
    # Was an input file given?
    if (defined $infile) {
        # Process a single user-specified file
        push (@infiles, $infile);
    } else {
        # Process all .m3m files in specified directory
        @infiles = sort(<*>);
    }
    
    # Iterate through the list of input files
    foreach my $infile (@infiles) {
        if ($infile =~ /\.m3m/) {
            open INFILE, "<", $infile || EatShitAndDie("Unable to read: $infile!");
            # First parse the file, and see what definitions are needed
            print "Processing: $infile....\n";
            
            # Pass 1 => Collect definitions, and expand macros
            # Slurp the file into an array, and a single string, to speed up search
            my $filetext = "";
            my @filelines = ();
            my %includes = ();
            my @includefiles = ();
            my %macros = ();
            $outfile = "$outdir/" . $infile;
            $outfile =~ s/\.m3m/.m1s/;
            my $infileline = 0;
            foreach (<INFILE>) {
                $infileline++;
                if (/^\s*#/) {
                    if (/^\s*#filename\s+\"([a-zA-Z][a-zA-Z0-9_\.]*)\"/) {
                        # Found output filename specification
                        $outfile = "$outdir/$1";
                        $filetext .= "\'$_";
                        push(@filelines, "\'$_");
                    } elsif (/^(\s*)#include\s+\"([a-zA-Z][a-zA-Z0-9_\.\\\:\& ]*)\"/) {
                        # Got an include file, so parse it
                        my $pad = $1;
                        my $includefile = $2;
                        $filetext .= "\'$_";
                        push(@filelines, "\'$_");
                        push (@includefiles, $includefile);
                        ParseInclude($includefile, \%includes, \%macros);
                    } elsif (/^(\s*)#([a-zA-Z][a-zA-Z0-9_]*)/) {
                        # Found a macro invocation, so expand it
                        my $pad = $1;
                        my $macroname = $2;
                        $filetext .= "$pad\' Macro: $_";
                        push(@filelines, "$pad\' Macro: $_");
                        # parse any macro arguments
                        $_ =~ s/^(\s*)#([a-zA-Z][a-zA-Z0-9_]*)\s*//;
                        $_ =~ s/\'.*$//;
                        my @macroargs = split(" ", $_);
                        while (@macroargs < 10) {
                            push (@macroargs, 0);
                        }
                        if (defined $macros{$macroname}) {
                            my @macrolines = split(/\n/, $macros{$macroname});
                            foreach my $macroline (@macrolines){
                                my $arg = 0;
                                foreach my $macroarg (@macroargs) {
                                    $macroline =~ s/\@($arg)/$macroarg/g;
                                    $arg++;
                                }
                                $filetext .= "$pad$macroline\n";
                                push(@filelines, "$pad$macroline\n");
                            }
                            $filetext .= "$pad\' End Macro: #$macroname\n";
                            push(@filelines, "$pad\' End Macro: #$macroname\n");
                        } else {
                            print "Warning: Undefined Macro: $macroname at line: $infileline of $infile\n";
                            $warnings++;
                        }
                    }
                } else {
                    # Just a normal line...
                    $filetext .= $_;
                    push(@filelines, $_);
                }
            }
            close INFILE;
            
            # Pass 2 => Search for all referenced defines
            @activedefines = ();
            foreach my $define (sort(keys(%includes))) {
                if ($filetext =~ /$define/) {
                    push(@activedefines, $define);
                    # print "Found: $define\n";
                }
            }
            
            # Pass 3 => Write the output file
            open OUTFILE, ">", $outfile || EatShitAndDie("Unable to open: $outfile!");
            # Remember where we came from
            print OUTFILE "''========================================================================================\n";
            print OUTFILE "'  Created From: $infile\n";
            print OUTFILE "'  Includes: \n";
            foreach (@includefiles) {
                print OUTFILE "'       $_\n";
            }
            print OUTFILE "''========================================================================================\n";
            print OUTFILE "\n";
            # First insert the required defines
            if (@activedefines) {
                print OUTFILE "''========================================================================================\n";
                print OUTFILE "''	 Included Definitions:\n";
                print OUTFILE "''========================================================================================\n";
                foreach (@activedefines) {
                    print OUTFILE $includes{$_};
                }
                print OUTFILE "''========================================================================================\n";
                print OUTFILE "\n";
            }
            # Now output the rest of the file
            foreach (@filelines) {
                    print OUTFILE $_;
            }
            close OUTFILE;
        }
    }
    print "Done!\n";
    print "$warnings warnings\n";


#------------------------------------------------------------------------------
#   ParseInclude
#------------------------------------------------------------------------------
sub ParseInclude {
    
    # Parse include file
    my $includefile = shift;
    my $includes = shift;
    my $macros = shift;
    $macros->{"filename"} = "";
    if (defined $includefile) {
        open FILE, "<", $includefile or EatShitAndDie("Unable to read: $includefile!");
        print "Parsing " . $includefile . "...\n";
        my $inmacro = 0;
        my $macroname = "";
        my $macrotext = "";
        foreach (<FILE>) {
            my $line = $_;
            if ($inmacro) {
                if (/^\#endm/) {
                    $macros->{$macroname} = $macrotext;
                    $inmacro = 0;
                    $macroname = "";
                    $macrotext = "";
                } else {
                    if (/^(\s*)#([a-zA-Z][a-zA-Z0-9_]*)/) {
                        foreach (split("\n", $macros->{$2})) {
                            $macrotext .= "$1$_\n";
                        }
                    } else {
                        $macrotext .= $_;
                    }
                }
            } else {
                if (/^\#macro\s+([a-zA-Z][a-zA-Z0-9_]*)/) {
                    # We've got a macro
                    $inmacro = 1;
                    $macroname = $1;
                    my $macrotext = "";
                } else {
                    # Just a normal line...
                    $_ =~ s/\s*//g;
                    if ($_ =~ /^[a-zA-Z].*\=/) {
                        my $define = $_;
                        $define =~ s/=.*$//;
                        $includes->{$define} = $line;
                    }
                }
            }
        }
    }
}    

#------------------------------------------------------------------------------
#   Help
#------------------------------------------------------------------------------
sub Help
{
    EatShitAndDie("No Help!");
}


#------------------------------------------------------------------------------
#   EatShitAndDie
#------------------------------------------------------------------------------
sub EatShitAndDie
{
    my $msg = shift;
    print "Error: $msg\n";
    print "Hit Enter to continue...\n";
    <STDIN>;
    die;
}

1;

